//-----------------------------------------------------------------------------
//
//	PICSTART Plus programming interface
//
//-----------------------------------------------------------------------------
//
//	Cosmodog, Ltd.
//	415 West Huron Street
//	Chicago, IL   60610
//	http://www.cosmodog.com
//
//-----------------------------------------------------------------------------
//
//	this interface to the PICSTART Plus programmer is provided in an effort
//	to make the PICSTART (and thus, the PIC family of microcontrollers) useful
//	and accessible across multiple platforms, not just one particularly well-
//	known, unstable one.
//
//-----------------------------------------------------------------------------
//
//	Copyright (C) 1999-2002 Cosmodog, Ltd.
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//-----------------------------------------------------------------------------
//
// revison history
//
//	0.2  (10 Sept. 1999) -- first real release
//
//	0.2a (11 Sept. 1999)
//		removed timebits.h from includes.h (wasn't actually used anywhere)
//
//	0.3  (12 Sept. 1999)
//		cleaned up a couple of mistakes in the usage text
//		added support for many more parts (everything supported by MPLAB)
//
//	0.4  (1 Oct. 1999)
//		cleaned up some argument handling (better, more generalized)
//		broke picstart.c up into smaller modules
//		added support for -wo flag (only works on single-word calibration spaces)
//		added oscillator calibration space size checking
//		added dump of device info if invoked without any flags
//
//	0.4a (28 Oct. 1999)
//		improved hex record parsing
//		added timeout to InitDevice() to allow the PICSTART to respond to DTR
//
//	0.4b (14 July 2000)
//		added support for several additional parts (culled from MPLAB 5.00.00)
//		added extended device profile (or something) to initialization -- newer
//			versions of MPLAB send more stuff at device init; these are now sent
//			by picp as well
//
//	0.4c (14 July, 2000)
//		added support for PIC18C442 and PIC18C452
//
//	0.4d (20 February, 2001)
//		changed CMD_REQUEST_ACK to CMD_REQUEST_MODEL (actually asking for programmer
//			model, not just an acknowledge)
//		dropped baud rates over 115200 from serial.c, they aren't useful here and
//			were limiting cross compatibility
//		added bool as a typedef so picp can be compiled under straight c instead of c++
//		changed Makefile so picp compiles using gcc instead of c++
//		modified OpenDevice() and CloseDevice() to save the previous serial port
//			settings and restore them on exit
//		made some improvements to ConfigureDevice() courtesy Brian Chamberlain
//
//	0.5 (31 October, 2002)
//		added support for more parts (current as of MPLAB 5.40.00 -- thanks Mike!)
//		increased character timeout to 1 second from 1/2 second (helps to eliminate
//			false "no response from PICSTART" errors)
//		added support for extended linear address records in Intel hex files.  this
//			used to fail silently and prevented programming if any byte address in
//			the file exceeded 0x8000 (like certain configuration words)
//		corrected size of 'startAddr' and 'curAddr' (were 16 bit, now need to be 32 bit)
//		added warning if segment address record is encountered in Intel hex files.
//			such records are now ignored; previously they caused the parser to fail
//			silently.
//		added test for buffer overrun when reading hex files.  this was absent before,
//			and writing files larger than 8K bytes would fail.  now hex files may be
//			arbitrarily large.
//		added extended linear address record generation to output when reading device
//		replaced some ioctl calls in serial.c with wrapper functions which makes it
//			more portable (now builds under OS X without modification) -- thanks to
//			Ahmi Wolf.
//		fixed error reporting in main() when you specify an illegal part number or
//			serial port
//		reorganized some of the code.
//		added support for V3 of the PS+.
//		added support for the 16F627.
//
//	0.5a (2 November 2002)
//		updated to match MPLAB 5.70.40 (PICSTART Plus firmware 3.00.40)
//		added all parts from MPLAB 5.70.40
//
//	0.5b (3 November 2002)
//		fixed bug which crept back into DoWritePgm() which messed up config word
//		increased BUFFERSIZE to 128k (was 8k)
//
//	0.5c (12 November 2002)
//		added signal handing; if any signal is received reset the PICSTART and exit
//		added libc++ to Makefile to eliminate link errors in OS X and Red Hat 8.0
//
//	0.5d (12 June 2003)
//		added erase flash command
//		fixed segfault in WritePgmRange()
//
//-----------------------------------------------------------------------------
//
// Special thanks to Todd Squires, Brian Chamberlain, Scott Johnston, Mike
//  Crowe, Jim Gasbarro, Tim Voght, Ahmi Wolf, and Antonio Augusto Todo Bom Neto
//  for various suggestions and contributions
//
//-----------------------------------------------------------------------------
//
//  to do:
//
//		read/write/erase ID locations
//		read/write/erase/verify data space
//		erase config bits
//		erase oscillator calibration (probably unnecessary since none of the current flash devices have calibration space)
//		verify program space
//		support osc. calibration space sizes other than 1 (PIC14000 only?)
//		seems to have trouble verifying oscillator calibration when stringing several operations together
//		seems to have a problem with 16C63A (but not 16C63, which MPLAB appears to treat as identical)
//		need to be able to write motorola S records
//
//-----------------------------------------------------------------------------

#include "includes.h"

#define CHAR_TIMEOUT		1000000			// time to wait for a character before giving up (in microseconds)
#define BUFFERSIZE			0x20000			// buffer bigger than anyone should need

#define MAXNAMESLEN			80				// max number of characters on a line when reporting device names


// COMMANDS (sent to PICSTART)
//
#define CMD_BLANK_CHECK		'B'				// blank check (0x0f = F not blank, 0x17 = not blank, 0x10 = blank?)
#define CMD_WRITE_PGM		'Q'				// write program memory
#define CMD_READ_PGM		'T'				// read program memory
#define CMD_READ_DATA		'd'				// read data (EEPROM) memory
#define CMD_READ_ID			'e'				// read ID locations
#define CMD_READ_CFG		'f'				// request configuration bits
#define CMD_READ_OSC		'c'				// read oscillator calibration memory
#define CMD_WRITE_CFG		'g'				// write configuration bits
#define CMD_WRITE_OSC		'q'				// write oscillator calibration memory
#define CMD_LOAD_INFO		0x81			// send processor-specific info
#define CMD_LOAD_EXT_INFO	0x82			// send more processor-specific info (this is new as of 0.4b)
#define CMD_REQUEST_MODEL	0x88			// request programmer model
#define CMD_REQUEST_VERSION	0x8d			// request firmware version
#define CMD_SET_ADDR		0x8e			// set address range (start address, size)
#define CMD_ERASE_FLASH		0x8f			// send erase flash device command

#define PIC_ACK				0xab			// response from picstart plus to model request (only programmer supported)


static const char versionString[] =	"0.5d";	// this program's version number

#define MIN_MAJORVER	3					// picstart version number detected must be at least this high
#define MIN_MIDVER		00
#define MIN_MINORVER	40

#define MINIMUM_VER		(MIN_MAJORVER*65536 + MIN_MIDVER*256 + MIN_MINORVER)

#define HASH_WIDTH_DEFAULT		40			// default width of the status bar

typedef struct
{
	unsigned char major;
	unsigned char middle;
	unsigned char minor;
} VERSION;

extern int
	errno;

static char
	*programName,
	*deviceName,
	*picName;

static VERSION
	PICversion;

static bool
	verboseOutput;
static bool
	ignoreVerfErr;
static int
	serialDevice;
static unsigned short int
	readConfigBits;							// config bits read back from device
static unsigned int
	hashWidth;								// width of status bar (0 = none)



//-----------------------------------------------------------------------------
// reset the PICSTART Plus by lowering then raising DTR
static void ResetPICSTART()
{
	SetDTR(serialDevice,false);				// lower DTR to reset PICSTART Plus
	usleep(1000000/4);						// sleep for a quarter second (to let PS+ reset)
	SetDTR(serialDevice,true);				// raise DTR
}


//-----------------------------------------------------------------------------
// process signals (any signal will cause us to exit)
static void SigHandler(int sig)
{
	fprintf(stderr,"exiting...\n");
	ResetPICSTART();
	exit(0);
}


//-----------------------------------------------------------------------------
// Find the requested PIC device from the list of supported devices.
// If the user didn't prepend 'PIC' to the device name skip the
// first three letters of the devices when comparing.  This allows
// both PIC16C505 or 16C505 to refer to the same part.
const static PIC_DEFINITION *GetPICDefinition(char *name)
{
	int
		idx;
	int
		offset;

	idx = 0;
	while(name[idx])						// ensure it's all upper case
	{
		name[idx] = toupper(name[idx]);
		idx++;
	}

	offset = 0;								// assume no PIC at the front
	if(strncmp(name,"PIC",3) == 0)			// if the user's argument has 'PIC' at the front,
	{
		offset = 3;							// skip the first three letters of the user's argument
	}

	idx = 0;
	while(deviceList[idx])
	{
		if(strcmp(deviceList[idx]->name,name+offset) == 0)
		{
			break;
		}
		idx++;
	}
	return(deviceList[idx]);		// return 0 if no match
}


//-----------------------------------------------------------------------------
// return the size of the program space of the specified device (in words)
static unsigned short int GetPgmSize(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_PGM_SIZEH]*256 + picDevice->def[PD_PGM_SIZEL]);
}


//-----------------------------------------------------------------------------
// return the size of the data space of the specified device (in bytes)
static unsigned short int GetDataSize(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_DATA_SIZEH]*256 + picDevice->def[PD_DATA_SIZEL]);
}


//-----------------------------------------------------------------------------
// return the start address of the data space of the specified device
static unsigned short int GetDataStart(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_DATA_ADDRH]*256 + picDevice->def[PD_DATA_ADDRL]);
}


//-----------------------------------------------------------------------------
// return the size of the oscillator calibration space of the specified device (in words)
static unsigned short int GetOscCalSize(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_CLK_SIZEH]*256 + picDevice->def[PD_CLK_SIZEL]);
}


//-----------------------------------------------------------------------------
// return the start address of the oscillator calibration space of the specified device
static unsigned short int GetOscCalStart(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_CLK_ADDRH]*256 + picDevice->def[PD_CLK_ADDRL]);
}


//-----------------------------------------------------------------------------
// return the start address of the configuration bits
static unsigned short int GetConfigStart(const PIC_DEFINITION *picDevice)
{
	return(picDevice->def[PD_CFG_ADDRH]*256 + picDevice->def[PD_CFG_ADDRL]);
}


//-----------------------------------------------------------------------------
//	send a message to the PICSTART, wait for a specified number of bytes to be returned
//  If false is returned, there was a timeout, or some other error
static bool SendMsg(const unsigned char *cmdBuff, unsigned int cmdBytes, unsigned char *rtnBuff, unsigned int rtnBytes)
{
	bool
		fail;
	int
		numRead;
	int
		bytesRemaining;

	fail=false;
	if(cmdBytes)
	{
		write(serialDevice,&cmdBuff[0],cmdBytes);								// send out the command
	}

	bytesRemaining = rtnBytes;
	while(bytesRemaining && !fail)
	{
		numRead = ReadBytes(serialDevice,&rtnBuff[rtnBytes-bytesRemaining],bytesRemaining,CHAR_TIMEOUT);

		if(numRead < 0)
		{
			fprintf(stderr,"error %d, %s\n",errno,strerror(errno));
			fail = true;
		}
		else if(numRead == 0)		// timed out
		{
			fail = true;
		}
		bytesRemaining = (bytesRemaining > numRead) ? bytesRemaining - numRead : 0;		// subtract bytes read in, don't allow to underflow (shouldn't happen)
	}
	return(!fail);
}


//-----------------------------------------------------------------------------
//	ask what type of programmer is attached (if any)
//
//  currently only the PICSTART Plus is supported; this is where MPLAB tries
//    to identify what type of programmer is attached and selects the
//    appropriate protocol based on the response.  Since only the PICSTART
//    Plus is supported we just check for its response code and fail if we
//    don't get it.
static bool DoGetProgrammerType()
{
	bool
		succeed;
	unsigned char
		theBuffer[1];
	unsigned int
		retryCount;

	theBuffer[0] = CMD_REQUEST_MODEL;
	retryCount=5;
	succeed=false;
	do
	{
		if(SendMsg(theBuffer,1,theBuffer,1))
		{
			if(theBuffer[0]==PIC_ACK)		// see if PICSTART Plus
			{
				succeed=true;
			}
			else
			{
				fprintf(stderr,"%02X\n",theBuffer[0]);
			}
		}
		retryCount--;
	} while(!succeed&&retryCount);

	return(succeed);
}


//-----------------------------------------------------------------------------
// request version from the PICSTART
static bool DoGetVersion()
{
	bool
		fail;
	unsigned char
		theBuffer[4];

	fail = false;
	theBuffer[0] = CMD_REQUEST_VERSION;
	if(SendMsg(theBuffer,1,theBuffer,4))
	{
		if(theBuffer[0] != CMD_REQUEST_VERSION)
		{
			fail = true;
		}
		else
		{
			PICversion.major = theBuffer[1];
			PICversion.middle = theBuffer[2];
			PICversion.minor = theBuffer[3];
		}
	}
	else
	{
		fprintf(stderr,"failed to read PICSTART version number\n");
		fail=true;
	}
	return(!fail);
}

//-----------------------------------------------------------------------------
//	get the version from the PICSTART and display it
static void DoShowVersion()
{
	fprintf(stdout,"PICSTART Plus firmware version %d.%02d.%02d\n",PICversion.major,PICversion.middle,PICversion.minor);
}

//-----------------------------------------------------------------------------
//  verify that the version number of the attached PICSTART is high enough to
//  work with this program, return false if it is not
static bool QualifyVersion(VERSION *PICversion)
{
	return( (PICversion->major*65536 + PICversion->middle*256 + PICversion->minor) >= MINIMUM_VER);	// true if >= min
}


//-----------------------------------------------------------------------------
// send a "set range" command to the PS+
static bool SetRange(unsigned int start,unsigned int length)
{
	unsigned char
		rangeBuffer[6],
		rtnBuffer[6];

	rangeBuffer[0] = CMD_SET_ADDR;
	rangeBuffer[1] = (start>>16) &0xFF;
	rangeBuffer[2] = (start>>8) & 0xFF;
	rangeBuffer[3] = start & 0xFF;
	rangeBuffer[4] = (length>>8) & 0xFF;
	rangeBuffer[5] = length & 0xFF;

	if(SendMsg(rangeBuffer,sizeof(rangeBuffer),rtnBuffer,sizeof(rtnBuffer)))
	{
		if(memcmp(rangeBuffer,rtnBuffer,sizeof(rangeBuffer))==0)	// read back result and see if it looks correct
		{
			return(true);
		}
		else
		{
			fprintf(stderr,"echoback from set range command incorrect\n");
		}
	}
	else
	{
		fprintf(stderr,"failed to send set range command\n");
	}
	return(false);
}


typedef unsigned short int SIZEFNCT(const PIC_DEFINITION *);

typedef struct
{
	unsigned char
		mask;
	SIZEFNCT
		*sizef;				// pointer to function to return the size of the space (0=no function)
	char
		*name;
} BLANK_MSG;


static const BLANK_MSG blankList[] =
{
	{BLANK_PGM,		&GetPgmSize,	"program memory"},
	{BLANK_CFG,		0,				"configuration bits"},
	{BLANK_ID,		0,				"ID locations"},
	{BLANK_DATA,	&GetDataSize,	"data memory"},
	{0,0,""},
};


//-----------------------------------------------------------------------------
// Check device for blank
static bool DoBlankCheck(const PIC_DEFINITION *picDevice, unsigned char blankMode)
{
	bool
		fail;
	unsigned char
		theBuffer[3];
	int
		idx;

	idx = 0;
	fail = false;
	theBuffer[0] = CMD_BLANK_CHECK;
	if(SendMsg(theBuffer,1,theBuffer,2))
	{
		theBuffer[1] &= blankMode;										// look only at what we were asked to look at
		if(!verboseOutput)
		{
			fprintf(stdout,"0x%02x\n",theBuffer[1]);					// quiet mode will just show the return code
		}
		else
		{
			while(blankList[idx].mask)									// report anything that isn't blank
			{
				if(blankMode & blankList[idx].mask)						// check only what we were asked
				{
					if(!blankList[idx].sizef || (blankList[idx].sizef(picDevice) > 0))
					{
						fprintf(stdout,"%s: ",blankList[idx].name);
						if(theBuffer[1] & blankList[idx].mask)			// if a bit is set,
						{
							fprintf(stdout,"not blank\n");
							fail = true;								// something wasn't blank
						}
						else
						{
							fprintf(stdout,"blank\n");
						}
					}
				}
				idx++;
			}
		}
	}
	else
	{
		fprintf(stderr,"failed to send blank check command\n");
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
#if 0		// this waits until we know how big the EEPROM space should be
static bool DoReadData(const PIC_DEFINITION *picDevice, FILE *theFile)
{
	bool
		fail;
	unsigned char
		theBuffer[?];

	fail = false;
	theBuffer[0] = CMD_READ_DATA;
	if(SendMsg(theBuffer,1,theBuffer,?))						// ask it to fill the buffer (plus the command plus a terminating zero)
	{
		WriteHexRecord(theFile,&theBuffer[1],0,size);			// write hex records to selected stream
	}
	else
	{
		fprintf(stderr,"failed to send read data command\n");
		fail=true;
	}
	return(!fail);
}
#endif


//--------------------------------------------------------------------
//  status bar handling
static unsigned short int
	hashMod,
	hashNum;

//--------------------------------------------------------------------
// initialize a status bar, given the number
// of operations expected
static void InitHashMark(unsigned short int numOps, unsigned short int hashWidth)
{
	hashNum = 0;
	if(hashWidth)							// if width = zero do nothing (no bar)
	{
		if(hashWidth <= numOps)				// if it will create less than 1 mark per operation
		{
			hashMod = numOps/hashWidth;		// mod is the number of operations divided by the bar width
		}
		else
		{
			hashMod = 1;					// don't allow the bar to be longer than the number of operations
		}
	}
	else
	{
		hashMod = 0;						// no bar, no mod
	}
}

//--------------------------------------------------------------------
// advance the status bar if the current number
// of operations warrants it
static void ShowHashMark(unsigned short int curOps)
{
	if(verboseOutput && hashMod && (curOps/hashMod > hashNum) )
	{
		fprintf(stdout,"#");
		fflush(stdout);						// be sure it gets displayed right away
		hashNum++;
	}
}


//--------------------------------------------------------------------
// finish up a status bar
static void UnInitHashMark()
{
	if(verboseOutput && hashMod)
	{
		fprintf(stdout,"\n");
	}
}


//--------------------------------------------------------------------
// read oscillator calibration
static bool DoReadOscCal(const PIC_DEFINITION *picDevice)
{
	bool
		fail;
	unsigned char
		*theBuffer;
	unsigned short int
		size;

	fail = false;
	size = GetOscCalSize(picDevice);
	if(size)
	{
		if(SetRange((picDevice->def[PD_CLK_ADDRH]<<8)+picDevice->def[PD_CLK_ADDRL],(picDevice->def[PD_CLK_SIZEH]<<8)+picDevice->def[PD_CLK_SIZEL]))
		{
			if((theBuffer = (unsigned char *)malloc(size+2)))					// get a buffer this big plus one char for the command and a 0 at the end
			{
				theBuffer[0] = CMD_READ_OSC;
				if(SendMsg(theBuffer,1,theBuffer,size+2))						// ask it to fill the buffer (plus the command plus a terminating zero)
				{
					int idx;
					fprintf(stdout,"oscillator calibration: ");
					for(idx = 1; idx<size+1; idx+=2)
					{
						if( (((idx-1)/2)&8) == 0 )
						{
							fprintf(stdout,"\n");
						}
						fprintf(stdout," 0x%02x%02x",theBuffer[idx],theBuffer[idx+1]);
					}
					fprintf(stdout,"\n");
				}
				else
				{
					fprintf(stderr,"failed to send read osc command\n");
					fail = true;
				}
			}
			else
			{
				fprintf(stderr,"failed to malloc\n");
				fail = true;
			}
		}
		else
		{
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"device %s has no oscillator calibration space\n",picDevice->name);
		fail = true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// write oscillator calibration
static bool DoWriteOscCalBits(const PIC_DEFINITION *picDevice, unsigned short int oscCalBits)
{
	bool
		fail;
	unsigned char
		theBuffer[3],
		rtnBuffer[4];
	unsigned short int
		size;								// size of the calibration space

	fail = false;
	size = GetOscCalSize(picDevice);		// size of the calibration space
	if(size == 1)
	{
		theBuffer[0] = CMD_WRITE_OSC;
		theBuffer[1] = (oscCalBits>>8) & 0xff;
		theBuffer[2] = oscCalBits & 0xff;
		if(SendMsg(theBuffer,3,rtnBuffer,4))
		{
			if(theBuffer[0] != rtnBuffer[0] || theBuffer[1] != rtnBuffer[1] || theBuffer[2] != rtnBuffer[2] || rtnBuffer[3] != 0)
			{
				fail = true;
				fprintf(stderr,"failed to verify while writing oscillator calibration data\n");
			}
		}
		else
		{
			fprintf(stderr,"failed to send write osc command\n");
			fail = true;
		}
	}
	else if(size == 0)
	{
		fprintf(stderr,"device %s has no oscillator calibration space\n",picDevice->name);
		fail = true;
	}
	else
	{
		fprintf(stderr,"oscillator calibration space sizes other than 1 not supported yet\n");
		fail = true;
	}
	return(!fail);
}

//--------------------------------------------------------------------
// read configuration bits
static bool DoReadCfg(bool verbose)
{
	bool
		fail;
	unsigned char
		theBuffer[4];

	fail = false;
	theBuffer[0] = CMD_READ_CFG;
	if(SendMsg(theBuffer,1,theBuffer,4))
	{
		if( (theBuffer[0] != CMD_READ_CFG) || (theBuffer[3] != 0) )
		{
			fprintf(stderr,"failed to read configuration bits\n");
			fail = true;
		}
		else
		{
			readConfigBits = theBuffer[1]*256 + theBuffer[2];
			if(verbose)
			{
				if(verboseOutput)
				{
					fprintf(stdout,"configuration bits: ");
				}
				fprintf(stdout,"0x%04x\n",readConfigBits);
			}
		}
	}
	else
	{
		fprintf(stderr,"failed to send read configuration command\n");
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// erase the program space of a part that can be erased (PIC16Fxx, etc)
static bool DoErasePgm(const PIC_DEFINITION *picDevice)
{
	bool
		fail;
	unsigned char
		theBuffer[4],
		rtnBuffer[2];
	unsigned short int
		size;								// size of the device's program memory (in bytes)
	int
		byteCnt;

	fail = false;
	size = GetPgmSize(picDevice)*2;			// get the size
	InitHashMark(size, hashWidth);

	if(SetRange(0,(picDevice->def[PD_PGM_SIZEH]<<8)+picDevice->def[PD_PGM_SIZEL]))	// erase the whole program space
	{
		theBuffer[0] = CMD_WRITE_PGM;
		if(SendMsg(theBuffer,1,rtnBuffer,1))				// send the command, watch for it to bounce back
		{
			if(*rtnBuffer == CMD_WRITE_PGM)
			{
				theBuffer[0] = 0xff;						// send as all 1's, even if device isn't 16 bits wide
				theBuffer[1] = 0xff;
				byteCnt = 0;
				while( (byteCnt < size) && !fail)
				{
					if(SendMsg(theBuffer,2,rtnBuffer,2))	// write the bytes, ignore the return value (check results later)
					{
						byteCnt+=2;
						ShowHashMark(byteCnt);
					}
					else
					{
						fprintf(stderr,"failed to send write program command\n");
						fail=true;
					}
				}
				UnInitHashMark();
				if(SendMsg(theBuffer,0,rtnBuffer,1))				// eat the trailing zero
				{
					if(*rtnBuffer == 0)
					{
						if(!DoBlankCheck(picDevice, BLANK_PGM))		// make sure it is now blank
						{
							fprintf(stderr,"failed to erase program space\n");
							fail=true;
						}
					}
					else
					{
						fprintf(stderr,"bad return result\n");
						fail = true;								// fail if it's not zero
					}
				}
				else
				{
					fprintf(stderr,"failed to read trailing 0\n");
					fail = true;									// didn't echo everthing back like it should have
				}
			}
			else
			{
				fprintf(stderr,"echoback did not look correct\n");
				fail = true;
			}
		}
		else
		{
			fprintf(stderr,"failed to send write program command\n");
			fail = true;										// didn't echo everthing back like it should have
		}
	}
	else
	{
		fail = true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// Execute a ERASE FLASH DEVICE operation
static bool DoEraseFlash(const PIC_DEFINITION *picDevice)
{
	bool
		fail;
	unsigned char
		theBuffer[2];

	fail=false;
	theBuffer[0] = CMD_ERASE_FLASH;
	if(SendMsg(theBuffer,1,theBuffer,2))
	{
		if((theBuffer[0]!=CMD_ERASE_FLASH)||(theBuffer[1]!=0)||(!DoBlankCheck(picDevice, BLANK_PGM|BLANK_CFG|BLANK_ID|BLANK_DATA)))	// make sure everything is now blank
		{
			fprintf(stderr,"failed to erase flash device\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"failed to send erase command\n");
		fail=true;
	}
	return(!fail);
}



//--------------------------------------------------------------------
// Write device's configuration bits
static bool DoWriteConfigBits(const PIC_DEFINITION *picDevice, unsigned short int configBits)
{
	bool
		fail;
	unsigned char
		theBuffer[3],
		rtnBuffer[3];

	fail = false;
	// DEBUG should do processor-specific range check here, if appropriate
	theBuffer[0] = CMD_WRITE_CFG;
	theBuffer[1] = (configBits>>8) & 0xff;
	theBuffer[2] = configBits & 0xff;
	if(SendMsg(theBuffer,3,rtnBuffer,3))
	{
		if(theBuffer[0] != rtnBuffer[0] || theBuffer[1] != rtnBuffer[1] || theBuffer[2] != rtnBuffer[2])
		{
			fprintf(stderr,"failed to verify while writing configuration bits\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"failed to send write configuration command\n");
		fail = true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
#if 0	// not done yet...
static bool DoWriteIDLocs(const PIC_DEFINITION *picDevice, unsigned short int configBits)
{
	bool
		fail;
	unsigned char
		theBuffer[3],
		rtnBuffer[3];

	fail = false;
	// DEBUG should do processor-specific range check here, if appropriate
	theBuffer[0] = CMD_WRITE_CFG;
	theBuffer[1] = (configBits>>8) & 0xff;
	theBuffer[2] = configBits & 0xff;
	if(SendMsg(theBuffer,3,rtnBuffer,3))
	{
		if(theBuffer[0] != rtnBuffer[0] || theBuffer[1] != rtnBuffer[1] || theBuffer[2] != rtnBuffer[2])
		{
			fprintf(stderr,"failed to verify while writing configuration bits\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"failed to send write configuration command\n");
		fail = true;
	}
	return(!fail);
}
#endif


//--------------------------------------------------------------------
// copy buffer to device, starting at word address startAddr_w, running for size_w words
//  DOES NOT boundary-check range -- will attempt to write outside of device's memory
//  Returns true if okay, false if failed
//  Verify error counts as failure only if failOnVerf = true
static bool WritePgmRange(unsigned short int startAddr_w, unsigned short int size_w, unsigned char *buffer)
{
	bool
		fail,
		verifyFail;
	unsigned char
		cmdBuffer[2];

	fail = verifyFail = false;

	if(SetRange(startAddr_w,size_w))
	{
		int idx = 0;

		while(idx < (size_w*2) )
		{
			unsigned char temp;
			temp = buffer[idx+1];
			buffer[idx+1] = buffer[idx];				// swap byte order (make it little endian)
			buffer[idx] = temp;
			idx += 2;
		}

		cmdBuffer[0] = CMD_WRITE_PGM;					// add in the command
		if(SendMsg(cmdBuffer,1,cmdBuffer,1))			// send the command, watch for it to bounce back
		{
			if(*cmdBuffer == CMD_WRITE_PGM)
			{
				idx = 0;
				while( !fail && (idx < (size_w*2)) )
				{
					if(SendMsg(&buffer[idx],2,cmdBuffer,2))
					{
						if( (buffer[idx] != cmdBuffer[0]) || (buffer[idx+1] != cmdBuffer[1]) )
						{
							verifyFail = true;						// didn't get back what we sent
						}
						idx+=2;
						ShowHashMark(idx);
					}
					else
					{
						fprintf(stderr,"failed to send write program command\n");
						fail = true;
					}
				}
				if(!fail)
				{
					if(SendMsg(buffer,0,cmdBuffer,1))					// eat the trailing zero
					{
						if(verifyFail)
						{
							if(!ignoreVerfErr)
							{
								fprintf(stderr,"failed to verify while writing to program space\n");
							}
							else
							{
								fprintf(stderr,"Warning: failed to verify while writing to program space\n");	// report it but don't fail on it
							}
						}
					}
					else
					{
						fprintf(stderr,"failed to get trailing 0\n");
						fail=true;
					}
				}
			}
			else
			{
				fprintf(stderr,"write program command did not echo back as expected\n");
				fail = true;
			}
		}
		else
		{
			fprintf(stderr,"failed to send write program command\n");
			fail = true;
		}
	}
	else
	{
		fail=true;
	}
	return( !(fail || (!ignoreVerfErr && verifyFail)) );
}


//--------------------------------------------------------------------
// write the program space of the passed device
static bool DoWritePgm(const PIC_DEFINITION *picDevice, FILE *theFile)
{
	bool
		fail,
		fileDone;
	unsigned char
		*theBuffer;
	unsigned int
		startAddr,		// hex record returns byte address, device expects word address -- always need to translate
		curAddr,
		nextAddr,
		size;
	unsigned char
		data;

	fail = fileDone = false;
	if((theBuffer = (unsigned char *)malloc(BUFFERSIZE)))
	{
		InitHashMark(picDevice->def[PD_PGM_SIZEH]*256+picDevice->def[PD_PGM_SIZEL], hashWidth);		// go to too much effort to set the width

		InitParse();														// get ready to start reading the hex file

		fileDone = !GetNextByte(theFile, &nextAddr, &data);					// get a byte and the initial address
		while(!fileDone && !fail)
		{
			startAddr = nextAddr;											// the first address of the new block
			curAddr = startAddr;
			theBuffer[0] = data;											// the first data byte of the new block
			size = 1;														// number of bytes waiting to be sent
			while( (!(fileDone = !GetNextByte(theFile, &nextAddr, &data)) &&  (size < BUFFERSIZE)) )	// get next byte
			{
				if(curAddr+1 == nextAddr)									// use this byte now only if it's sequential
				{
					curAddr++;
					theBuffer[size] = data;
					size++;
				}
				else
				{
					break;													// non-sequential, write this buffer then start another
				}
			}
			if(size & 1)
			{
				size++;														// don't allow odd sizes
				theBuffer[size] = 0xff;										// pad with 0xff (erased)
			}

			{
				// DEBUG check if config address is embedded in range
				//
				unsigned short int
					devCfgAddr = picDevice->def[PD_CFG_ADDRH]*256 + picDevice->def[PD_CFG_ADDRL];		// word address of device's config memory

				if( startAddr/2 == devCfgAddr)															// at config bits address?
				{
					if(size == 2)																		// must be exactly one word
					{																					// inside configuration space, write as configuration
						fail = !DoWriteConfigBits(picDevice, theBuffer[0] + theBuffer[1]*256);			// (little endian, data already swapped)
					}
					else
					{
						fprintf(stderr,"configuration word must be exactly 16 bits\n");
					}
				}
				else
				{
					fail = !WritePgmRange(startAddr/2, size/2, theBuffer);	// if not config bits try to write where ever it lands
				}
			}
		}
		UnInitHashMark();
		free(theBuffer);
	}
	else
	{
		fprintf(stderr,"failed to malloc %d bytes\n",BUFFERSIZE);
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// Read the program space of the passed device
static bool DoReadPgm(const PIC_DEFINITION *picDevice, FILE *theFile)
{
	bool
		fail;
	unsigned char
		*theBuffer;
	unsigned int
		size;						// size of the device's program memory (in bytes)

	fail = false;
	size = GetPgmSize(picDevice)*2;
	if(DoReadCfg(false))
	{
		if(~readConfigBits & picDevice->cpbits)
		{
			fprintf(stderr,"Warning: device is code protected: configuration bits = 0x%04x\n",readConfigBits);
		}
		if(SetRange(0,(picDevice->def[PD_PGM_SIZEH]<<8)+picDevice->def[PD_PGM_SIZEL]))
		{
			if((theBuffer = (unsigned char *)malloc(size+2)))		// get a buffer this big plus one char for the command and a 0 at the end
			{
				theBuffer[0] = CMD_READ_PGM;
				if(SendMsg(theBuffer,1,theBuffer,size+2))			// ask it to fill the buffer (plus the command plus a terminating zero)
				{
					// DEBUG shouldn't need to swap byte order here but we do

					unsigned int
						idx;
					unsigned char
						temp;

					for(idx = 1; idx<size+1; idx+=2)
					{
						temp = theBuffer[idx+1];
						theBuffer[idx+1] = theBuffer[idx];			// swap byte order (make it little endian)
						theBuffer[idx] = temp;
					}

					WriteHexRecord(theFile,&theBuffer[1],0,size);	// write hex records to selected stream
				}
				else
				{
					fprintf(stderr,"failed to send read program command\n");
					fail = true;
				}
				free(theBuffer);
			}
			else
			{
				fprintf(stderr,"failed to allocate buffer\n");
				fail = true;				// failed to malloc
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		fprintf(stderr,"failed to read config bits\n");
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// Read the device ID locations
static bool DoReadID()
{
	bool
		fail;
	unsigned char
		theBuffer[10];

	fail = false;
	theBuffer[0] = CMD_READ_ID;
	if(SendMsg(theBuffer,1,theBuffer,10))
	{
		if( (theBuffer[0] == CMD_READ_ID) && (theBuffer[9] == 0) )
		{
			if(verboseOutput)
			{
				fprintf(stdout,"ID locations: ");							// if in quiet mode, only the values will be returned
			}
			fprintf(stdout,"0x%02x%02x ",theBuffer[1],theBuffer[2]);
			fprintf(stdout,"0x%02x%02x ",theBuffer[3],theBuffer[4]);
			fprintf(stdout,"0x%02x%02x ",theBuffer[5],theBuffer[6]);
			fprintf(stdout,"0x%02x%02x",theBuffer[7],theBuffer[8]);
			fprintf(stdout,"\n");
		}
		else
		{
			fprintf(stderr,"failed to read ID locations\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"failed to send read ID command\n");
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// initialize for specified part, return true if succeeded
static bool DoInitPIC(const PIC_DEFINITION *picDevice)
{
	bool
		fail;
	unsigned char
		theBuffer[3];
	unsigned char
		cmdBuffer[PICDEV_DEFSIZE+1];
	unsigned char
		extCmdBuffer[PICDEV_DEFXSIZE+1];

	fail=false;
	theBuffer[0] = CMD_LOAD_INFO;
	if(SendMsg(theBuffer,1,theBuffer,1))									// send load processor info command, wait for command to echo back
	{
		if(theBuffer[0] == CMD_LOAD_INFO)
		{
			int idx;
			memcpy(cmdBuffer,picDevice->def,PICDEV_DEFSIZE);			// copy definition into the new buffer
			cmdBuffer[PICDEV_DEFSIZE] = 0;								// initialize the checksum
			for(idx = 0; idx < PICDEV_DEFSIZE; idx++)					// calculate the checksum, store as last byte in buffer
			{
				cmdBuffer[PICDEV_DEFSIZE] += cmdBuffer[idx];
			}
			if(SendMsg(cmdBuffer,PICDEV_DEFSIZE+1,theBuffer,1))				// send whole buffer including checksum
			{
				if(theBuffer[0] == 0)										// zero = checksum okay (data received okay)
				{
					theBuffer[0] = CMD_LOAD_EXT_INFO;
					if(SendMsg(theBuffer,1,theBuffer,1))					// send load extended processor info command, wait for command to echo back
					{
						if(theBuffer[0]==CMD_LOAD_EXT_INFO)
						{
							int idx;
							memcpy(extCmdBuffer,picDevice->defx,PICDEV_DEFXSIZE);	// copy definition into the new buffer
							extCmdBuffer[PICDEV_DEFXSIZE] = 0;						// initialize the checksum
							for(idx = 0; idx < PICDEV_DEFXSIZE; idx++)				// calculate the checksum, store as last byte in buffer
							{
								extCmdBuffer[PICDEV_DEFXSIZE] += extCmdBuffer[idx];
							}
							if(SendMsg(extCmdBuffer,PICDEV_DEFXSIZE+1,theBuffer,1))	// send whole buffer including checksum
							{
								if(theBuffer[0] != 0)								// zero = checksum okay (data received okay)
								{
									fprintf(stderr,"bad result from PICSTART\n");
									fail = true;									// didn't receive zero, fail
								}
							}
							else
							{
								fprintf(stderr,"failed to send extended device data\n");
								fail = true;
							}
						}
						else
						{
							fprintf(stderr,"echoback did not look correct\n");
							fail = true;
						}
					}
					else
					{
						fprintf(stderr,"failed to send extended device definition\n");
						fail = true;
					}
				}
				else
				{
					fprintf(stderr,"checksum failure\n");
					fail = true;											// didn't receive zero, fail
				}
			}
			else
			{
				fprintf(stderr,"failed to send device definition\n");
				fail = true;
			}
		}
		else
		{
			fprintf(stderr,"echoback did not look correct\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"failed to send load info command\n");
		fail=true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// find out if the next argument is a flag (not preceeded by '-').
// 	if it is, return a pointer to it and advance to the next argument, otherwise
//	return null.
static char *GetNextFlag(int *argc, char **argv[])
{
	if(*argc  && (***argv != '-'))						// if the next argument isn't preceeded by a '-'
	{
		(*argv)++;										// advance to next argument
		(*argc)--;
		return(*(*argv-1));								// but return pointer to this one
	}
	else
	{
		return(NULL);
	}
}

//--------------------------------------------------------------------
// Do all the things that the command line is asking us to do
static bool DoTasks(int *argc,char **argv[],const PIC_DEFINITION *picDevice,char *flags)
{
	bool fail = false;
	char *fileName = (char *)0;
	FILE *theFile;
	unsigned char blankMode;


	switch(*flags)
	{
		case 'b':								// blank check
			flags++;
			if(*flags)
			{
				blankMode = 0;
				while(*flags)
				{
					switch(*flags)
					{
						case 'p':
							blankMode |= BLANK_PGM;
							break;

						case 'c':
							blankMode |= BLANK_CFG;
							break;

						case 'i':
							blankMode |= BLANK_ID;
							break;

						case 'd':
							blankMode |= BLANK_DATA;
							break;

						default:
							break;				// ignore undefined flags

					}
					flags++;
				}
			}
			else
			{
				blankMode = BLANK_PGM | BLANK_CFG | BLANK_ID | BLANK_DATA;	// no mode flags means check them all
			}
			fail = !DoBlankCheck(picDevice,blankMode);
			break;

		case 'e':
			flags++;
			if(*flags)
			{
				while(*flags && !fail)
				{
					switch(*flags)
					{
						case 'p':
							fail = !DoErasePgm(picDevice);
							break;

						case 'c':
							fprintf(stderr,"erase config bits not implemented yet\n");
							break;

						case 'i':
							fprintf(stderr,"erase ID locations not implemented yet\n");
							break;

						case 'd':
							fprintf(stderr,"erase data memory not implemented yet\n");
							break;

						case 'o':
							fprintf(stderr,"erase oscillator calibration not implemented yet\n");	// DEBUG can osc cal be erased/written on flash parts?
							break;

						case 'f':
							fail = !DoEraseFlash(picDevice);
							break;

					}
					flags++;
				}
			}
			else
			{
				fprintf(stderr,"specify one or more regions to erase (p|c|i|d|o)\n");
			}
			break;

		case 'r':
			flags++;
			if(*flags)
			{
				while(*flags && !fail)
				{
					switch(*flags)
					{
						case 'p':
							if( (fileName = GetNextFlag(argc,argv)) )
							{
								theFile = fopen(fileName,"w");
							}
							else
							{
								theFile = stdout;
							}
							if(theFile)
							{
								fail = !DoReadPgm(picDevice, theFile);		// read program data, write to stream
								if(theFile != stdout)						// if we wrote it to a file,
								{
									fclose(theFile);						// close the file
								}
							}
							else
							{
								fprintf(stderr,"unable to open output file: '%s'\n",fileName);
							}
							break;

						case 'c':
							fail = !DoReadCfg(true);		// read configuration bits, display them
							break;

						case 'i':
							fail = !DoReadID();				// read ID locations
							break;

						case 'd':
							fprintf(stderr,"reading data memory not implemented yet\n");
							// read data memory
							break;

						case 'o':
							fail = !DoReadOscCal(picDevice);
							break;

					}
					flags++;
				}
			}
			else
			{
				fprintf(stderr,"specify one or more regions to read (p|c|i|d|o)\n");
			}
			break;

		case 'w':
			flags++;
			if(*flags)
			{
				switch(*flags)
				{
					case 'p':
						if( (fileName = GetNextFlag(argc,argv)) )
						{
							theFile = fopen(fileName,"r");
						}
						else
						{
							theFile = stdin;
						}
						if(theFile)
						{
							fail = !DoWritePgm(picDevice, theFile);
							if(theFile != stdin)						// if we read it from a file,
							{
								fclose(theFile);						// close the file
							}
						}
						else
						{
							fprintf(stderr,"unable to open input file: '%s'\n",fileName);
						}
						break;

					case 'c':
						if( (fileName = GetNextFlag(argc,argv)) )	// 'fileName' is actually the next argument
						{
							unsigned int configBits;
							fail = !atoi_base(fileName,&configBits);
							if(!fail)
							{
								if(configBits < 0x10000)
								{
									fail = !DoWriteConfigBits(picDevice, (unsigned short int)configBits);
								}
								else
								{
									fprintf(stderr,"Value out of range: '%s'\n",fileName);
									fail = true;
								}

							}
							else
							{
								fprintf(stderr,"Unable to interpret '%s' as a numerical value\n",fileName);
							}
						}
						else
						{
							fprintf(stderr,"write configuration bits must be followed by a numerical value\n");
						}
						break;

					case 'i':			// if writing ID locations, the next four arguments must be numbers
						fprintf(stderr,"writing ID locations not implemented yet\n");
						break;

					case 'd':
						fprintf(stderr,"writing data memory not implemented yet\n");
						break;

					case 'o':
						if( (fileName = GetNextFlag(argc,argv)) )	// 'fileName' is actually the next argument
						{
							unsigned int oscCalBits;
							fail = !atoi_base(fileName,&oscCalBits);
							if(!fail)
							{
								if(oscCalBits < 0x10000)
								{
									fail = !DoWriteOscCalBits(picDevice, (unsigned short int)oscCalBits);
								}
								else
								{
									fprintf(stderr,"Value out of range: '%s'\n",fileName);
									fail = true;
								}

							}
							else
							{
								fprintf(stderr,"Unable to interpret '%s' as a numerical value\n",fileName);
							}
						}
						else
						{
							fprintf(stderr,"write oscillator calibration must be followed by a numerical value\n");
						}
						break;

					default:
						fprintf(stderr,"must specify a region to write\n");
						break;

				}
			}
			else
			{
				fprintf(stderr,"specify a region to write (p|c|i|d|o)\n");
			}
			break;

		case 'v':
			fprintf(stderr,"DEBUG verify is not implemented yet\n");
			// DEBUG verify device goes here
			break;

		default:
			break;

	}
	return(!fail);
}

//--------------------------------------------------------------------
// Initialize the serial port
// Once the device is opened and locked, this sets up the port, and makes sure the handshake looks good.
static bool InitDevice(int serialDevice,unsigned int baudRate,unsigned char dataBits,unsigned char stopBits,unsigned char parity)
{
	bool
		fail;											// haven't failed (yet)
	bool
		CTS,
		DCD;
	unsigned short int
		ctsTimeOut;

	fail = false;
	if(ConfigureDevice(serialDevice,baudRate,dataBits,stopBits,parity,false))	// set up the device
	{
		if(ConfigureFlowControl(serialDevice,false))	// no flow control at the moment (raise RTS)
		{
			ResetPICSTART();
			ctsTimeOut = 100;							// allow about 100 ms (0.1 sec) for CTS to show up
			do
			{
				GetDeviceStatus(serialDevice,&CTS,&DCD);	// see if CTS is true
				if(CTS)
				{
					break;								// break out if it is
				}
				usleep(1000);							// wait 1 ms (more or less), try again
			}
			while(ctsTimeOut--);

			if(!CTS)
			{
				fprintf(stderr,"programmer not detected (CTS is false)\n");
				fail = true;							// didn't see CTS, assume device is not present or not ready, fail
			}
			else
			{
				ConfigureFlowControl(serialDevice,true);	// looks ok to use flow control, so allow it
			}
			FlushBytes(serialDevice);						// get rid of any pending data
		}
		else
		{
			fprintf(stderr,"could not configure flow control\n");
			fail = true;
		}
	}
	else
	{
		fprintf(stderr,"could not configure device parameters\n");
		fail = true;
	}
	return(!fail);
}


//--------------------------------------------------------------------
// Show a starting address and size
static void ShowStartSize(unsigned short int start, unsigned short int size)
{
	if(size)
	{
		fprintf(stdout,"    0x%04x-0x%04x (0x%04x word%c)\n",start,start+size-1,size,((size != 1)?'s':'\0'));
	}
	else
	{
		fprintf(stdout,"    none\n");
	}
}

//--------------------------------------------------------------------
// Report information about the passed device
static void ShowDeviceInfo(const PIC_DEFINITION *picDevice)
{
	fprintf(stdout,"device name: %s\n",picDevice->name);				// show the name

	fprintf(stdout,"  program space:\n");								// show range of program space
	ShowStartSize(0,GetPgmSize(picDevice));

	fprintf(stdout,"  data space:\n");									// show range of data space, if any
	ShowStartSize(GetDataStart(picDevice),GetDataSize(picDevice));

	fprintf(stdout,"  oscillator calibration space:\n");				// show range of calibration space, if any
	ShowStartSize(GetOscCalStart(picDevice),GetOscCalSize(picDevice));

	fprintf(stdout,"  configuration bits:\n");
	fprintf(stdout,"    address: 0x%04x\n",GetConfigStart(picDevice));	// show address of configuration bits
	fprintf(stdout,"    protect mask:  0x%04x\n",picDevice->cpbits);	// mask of code protect bits
	fprintf(stdout,"    watchdog mask: 0x%04x\n",picDevice->wdbit);		// mask of watchdog enable bit
}


//--------------------------------------------------------------------
//	display all supported devices
static void ShowDevices()
{
	int idx = 0;
	int length = 0;
	int thisLength;

	fprintf(stdout,"supported devices:\n");
	while(deviceList[idx])
	{
		thisLength = strlen(deviceList[idx]->name);			// length of this device's name
		length += thisLength;								// add to length of this line
		if(length+2 < MAXNAMESLEN)							// ensure there's room for the space and comma, too
		{
			fprintf(stdout,"%s",deviceList[idx]->name);		// put it on this line
		}
		else
		{
			fprintf(stdout,"\n%s",deviceList[idx]->name);	// put it on the next line
			length = thisLength;
		}
		idx++;
		if(deviceList[idx])									// if more devices are in the list,
		{
			fprintf(stdout,", ");							// add a comma and a space
			length+=2;
		}
		else
		{
			fprintf(stdout,"\n");							// or newline if it is the last one
		}
	}
}


//--------------------------------------------------------------------
// some thorny issues still exist in specifying arguments:
//
//	1) should write cause a blank check before writing?  should it program anyway if no bits that should be 1 are 0?
//	2) several different writes are needed (program, ID, data, config)
//	3) likewise, several different reads are needed
//	4) need to be able to specify a range of addresses to read, instead of reading the entire device
//	5) should erase be an option?  several different erases? erase plus a set of flags to indicate what to erase?

//--------------------------------------------------------------------
// tell the user how to use this program
static void Usage()
{
	fprintf(stdout,"%s: version %s, (c) 2000-2003 Cosmodog, Ltd. (http://www.cosmodog.com)\n",programName,versionString);
	fprintf(stdout,"Usage: %s ttyname devtype [-h] [-q] [-v] [-s [size]] [-b|-r|-w|-e][pcidof]\n",programName);
	fprintf(stdout," where:\n");
	fprintf(stdout,"  ttyname is the serial device the PICSTART is attached to (e.g., /dev/ttyS0)\n");
	fprintf(stdout,"  devtype is the pic device to be used (12C508, 16C505, etc.)\n");
	fprintf(stdout,"  -h shows this help\n");
	fprintf(stdout,"  -s [size] shows a hash mark status bar of length [size] while erasing/writing\n");
	fprintf(stdout,"  -q sets quiet mode (excess messages supressed)\n");
	fprintf(stdout,"  -r initiates a read (Intel Hex record format)\n");
	fprintf(stdout,"  -b blank checks the requested region or regions\n");
	fprintf(stdout,"  -f ignores verify errors while writing\n");
	fprintf(stdout,"  -w writes to the requested region\n");
	fprintf(stdout,"  -e erases the requested region (flash parts only)\n");
	fprintf(stdout,"  -v shows PICSTART Plus version number\n");
	fprintf(stdout,"    p [filename] = program memory, optionally reading/writing filename\n");
	fprintf(stdout,"    c [val] = configuration bits (val is a numeric word value when writing)\n");
	fprintf(stdout,"    i [val] = ID locations\n");
	fprintf(stdout,"    d [filename] = data memory, optionally reading/writing filename\n");
	fprintf(stdout,"    o [val] = oscillator calibration space\n");
	fprintf(stdout,"    f = entire flash device (only applies to -e, erase)\n");
	fprintf(stdout,"  filename is an optional input or output file (default is stdin/stdout)\n");
	fprintf(stdout,"\n");
	fprintf(stdout,"Flags are operated on in order, from left to right.  If any operation fails,\n");
	fprintf(stdout,"further execution is aborted.  Thus, a part can be blank checked and programmed\n");
	fprintf(stdout,"with a single command, e.g.:\n");
	fprintf(stdout,"        %s /dev/ttyS0 16c505 -bp -wp program.hex \n",programName);
	fprintf(stdout,"This example will blank check the program memory of a PIC16C505 then write the\n");
	fprintf(stdout,"contents of the file program.hex to the program memory only if the blank check\n");
	fprintf(stdout,"succeeded.\n");
	fprintf(stdout,"The -wc, -wi, and -wo options must be followed by a numeric argument which\n");
	fprintf(stdout,"represents the value.  The number may be binary (preceeded by 0b or 0B), hex\n");
	fprintf(stdout,"(preceeded by 0x or 0X), or decimal (anything else).\n\n");
	ShowDevices();
}


//--------------------------------------------------------------------
// Program PICs through a serial port
int main(int argc,char *argv[])
{
	bool
		fail;
	unsigned int
		baudRate;
	unsigned char
		dataBits,
		stopBits,
		parity;
	bool
		done;

	char
		*flags;

	const PIC_DEFINITION
		*picDevice=0;

	signal(SIGINT, SigHandler);								// set up a signal handler

	programName=*argv++;									// name of the application
	argc--;

	fail = false;
	verboseOutput=true;										// be verbose unless told otherwise
	hashWidth=false;										// don't show hask marks by default
	ignoreVerfErr=false;									// by default stop on verify errors

	if(argc>2)												// need at least four arguments to do anything
	{
		deviceName=*argv++;									// name of the device (probably)
		argc--;
		picName=*argv++;									// name of the PIC type (probably)
		argc--;
		if( (picDevice=GetPICDefinition(picName)) )			// locate the PIC type (0 = none found)
		{
			done=false;
			if(OpenDevice(deviceName,&serialDevice))		// open the serial device
			{
				baudRate=19200;
				dataBits=8;
				stopBits=1;
				parity=0;

				if(InitDevice(serialDevice,baudRate,dataBits,stopBits,parity))	// initialize the serial port
				{
					if(DoGetProgrammerType())									// ask what kind of programmer is attached, fail if none or one we don't support
					{
						if(DoGetVersion())										// get the version number of the PICSTART
						{
							if(QualifyVersion(&PICversion))						// be sure it's correct
							{
								if(DoInitPIC(picDevice))						// try to load up the parameters for this device
								{
									while(argc && !fail)						// do as long as we can read some more
									{
										flags = *argv++;						// get this argument, point to the next
										argc--;
										if(*flags == '-')						// see if it's a flag
										{
											flags++;							// it is, skip the dash
											switch(*flags)
											{
												case 'v':
													DoShowVersion();
													break;

												case 'f':
													ignoreVerfErr = true;		// force, ignore verify error
													break;

												case 'q':
													verboseOutput=false;		// inhibit display of messages
													break;

												case 'h':
													Usage();					// give help
													break;

												case 's':
													if(argc && **argv != '-')					// if the next argument isn't preceeded by a '-'
													{
														fail = !atoi_base(*argv,&hashWidth);	// try to read the next argument as a number
														argv++;									// skip to the next argument
														argc--;
														if(fail)
														{
															fprintf(stderr,"Unable to interpret '%s' as a numerical value\n",*argv);
														}
													}
													else
													{
														hashWidth = HASH_WIDTH_DEFAULT;			// turn hash marks on to the default width
													}
													break;

												case 'b':
												case 'r':
												case 'w':
												case 'e':
													fail = !DoTasks(&argc,&argv,picDevice,flags);	// do the requested operation
													break;

												case '\0':						// ignore a stray dash
													break;

												default:
													fprintf(stderr,"bad argument: '%s'\n",*(argv-1));			// back up, show the trouble spot
													break;

											}
										}
									}
								}
								else
								{
									fprintf(stderr,"failed to initialize %s\n",picDevice->name);
								}
							}
							else
							{
								DoShowVersion();
								fprintf(stderr,"firmware version must be at least %d.%02d.%02d\ngo to http://www.microchip.com to upgrade\n",MIN_MAJORVER,MIN_MIDVER,MIN_MINORVER);
							}
						}
						else
						{
							fprintf(stderr,"failed to obtain PICSTART firmware version number\n");
						}
					}
					else
					{
						fprintf(stderr,"failed to connect to PICSTART Plus\n");
					}
				}
				else
				{
					fprintf(stderr,"failed to set up the serial port\n");
				}
				CloseDevice(serialDevice);
			}
			else
			{
				fprintf(stderr,"failed to open device '%s'\n",deviceName);
			}
		}
		else
		{
			fprintf(stderr,"unrecognized PIC device type: '%s'\n",picName);		// don't know that one;
			ShowDevices();														// give a helpful list of supported devices
		}
	}
	else
	{
		if( (argc == 1 && (picDevice = GetPICDefinition(argv[0]))) ||			// locate the PIC type (0 = none found)
			(argc == 2 && (picDevice = GetPICDefinition(argv[1]))) )			//  (be forgiving about arg position)
		{
			ShowDeviceInfo(picDevice);
		}
		else
		{
			Usage();
			fail = true;
		}
	}
	return(fail);	// return 0 if okay (not failed)
}
